-- syntax
module Abstract where

import Util

import Data.Map (Map)
import qualified Data.Map as Map

type Name = String

data Sized = Sized | NotSized 
             deriving (Eq,Show)

data Co = Ind 
        | CoInd
          deriving (Eq,Show)

-- strict positivity 
data Pos 
  = SPos  -- strictly positive
  | NSPos -- unknown polarity
  | Neg   -- negative, used internally for contravariance of sized codata  
  | NOcc  -- non-occurring, used to determine whether data type is recursive
         deriving (Eq)

instance Show Pos where
  show SPos  = "+"
  show NSPos = "o"
  show Neg   = "-"
  show NOcc  = "T"

isSPos :: Pos -> Bool
isSPos SPos = True
isSPos NOcc = True
isSPos _ = False

polNeg :: Pos -> Pos
polNeg SPos = Neg
polNeg Neg = SPos
polNeg NSPos = NSPos
polNeg NOcc  = NOcc

-- polarity composition
polComp :: Pos -> Pos -> Pos
polComp NOcc  x   = NOcc
polComp x NOcc    = NOcc
polComp SPos  x   = x
polComp NSPos x   = NSPos
polComp Neg Neg   = SPos
polComp Neg NSPos = NSPos
polComp Neg SPos  = Neg

-- polarity conjuction (infimum)
polAnd :: Pos -> Pos -> Pos
polAnd NOcc x   = x
polAnd x NOcc   = x
polAnd NSPos x  = NSPos
polAnd x NSPos  = NSPos
polAnd SPos Neg = NSPos
polAnd Neg SPos = NSPos
polAnd SPos SPos = SPos
polAnd Neg Neg   = Neg

-- decoration of Pi-types
-- 1. whether argument is erased
-- further possibilities:
-- 2. polarity
-- 3. hidden

data Dec = Dec { erased :: Bool }
           deriving Eq

defaultDec = notErased
notErased = Dec False
argErased    = Dec True

instance Show Dec where
    show dec = if erased dec then "[-]" else "(-)"

type MVar = Int -- metavariables are numbered

data Expr = Set
          -- size type
          | Size 
          | Succ Expr
          | Infty
          -- identifiers 
          | Meta MVar    -- meta-variable
          | Var Name     -- variables are named
          | Con Co Name  -- constructors
          | Def Name     -- fun/cofun ?
          | Let Name     -- definition (non-recursive)
          -- dependently typed lambda calculus
          | Case Expr [Clause]
          | LLet Name Expr Expr Expr --local let
          | App Expr [Expr]
          | Lam Name Expr
          | Pi Dec Name Expr Expr
          deriving (Eq)

-- get all the definitions that are refered to in expression
-- this is used e.g. to check whether a (co)fun is recursive
usedDefs :: Expr -> [Name]
usedDefs (Def n) = [n]
usedDefs (App e es) = foldl (++) (usedDefs e) (map usedDefs es)
usedDefs (Lam x e)  = usedDefs e
usedDefs (Pi _ _ a b) = usedDefs a ++ usedDefs b
usedDefs (LLet _ e e1 e2) = usedDefs e ++ usedDefs e1 ++ usedDefs e2
usedDefs (Succ e) = usedDefs e
usedDefs _ = []


instance Show Expr where
    show = prettyExpr

prettyExpr :: Expr -> String
prettyExpr e = 
    case e of
      Set -> "Set"
      Size -> "Size" 
      Succ e -> "($ " ++ prettyExpr e ++ ")"
      Infty -> "#"
      Meta i -> "?" ++ show i
      Var n -> n
      Con _ n -> n
      Def n -> n
      Let n -> n
      Case e cs -> "case " ++ prettyExpr e ++ " { " ++ prettyCases cs ++ " } "
      LLet n t1 e1 e2 ->
          "(let " ++ n ++ " : " ++ prettyExpr t1 ++ " = " ++ prettyExpr e1 ++ " in " ++ prettyExpr e2 ++ ")"  
      App e1 el -> "(" ++ prettyExprs (e1:el) ++ ")"
      Lam x e1 -> "(\\" ++ x ++ " -> " ++ prettyExpr e1 ++ ")"
      Pi dec x t1 t2 -> parens $ (if erased dec then brackets binding
                                    else if x=="" then s1 else parens binding) 
                                  ++ " -> " ++ prettyExpr t2
         where s1 = prettyExpr t1
               binding = if x == "" then s1 else x ++ " : " ++ s1 
               
prettyCase (Clause [p] e) = prettyPattern p ++ " -> " ++ prettyExpr e
prettyCases = prettyList "; " prettyCase 

prettyPattern :: Pattern -> String
prettyPattern = prettyExpr . patternToExpr
{-
prettyPattern (ConP c _ ps) = parens $ foldl (\ acc p -> acc ++ " " ++ prettyPattern p) c ps
prettyPattern (SuccP p) = parens $ "$ " ++ prettyPattern p
prettyPattern (DotP e)  = "." ++ prettyExpr e
prettyPattern (VarP x) = x
-}

prettyExprs :: [Expr] -> String
prettyExprs = prettyList " " prettyExpr


-- parallel substitution (CAUTION! NOT CAPTURE AVOIDING!) 

parSubst :: (Name -> Expr) -> Expr -> Expr
parSubst phi (Var x) = phi x
parSubst phi (Succ e) = Succ (parSubst phi e)
parSubst phi (App e es) = App (parSubst phi e) (map (parSubst phi) es)
parSubst phi (LLet x a b c) = LLet x (parSubst phi a) (parSubst phi b) (parSubst phi c)
parSubst phi (Pi dec x a b) = Pi dec x (parSubst phi a) (parSubst phi b)
parSubst phi (Lam x e)  = Lam x (parSubst phi e)
parSubst phi e = e

sgSubst :: Name -> Expr -> Expr -> Expr
sgSubst x t u = parSubst (\ y -> if x == y then t else Var y) u

-- metavariable substitution (CAUTION! NOT CAPTURE AVOIDING!) 

type Subst = Map MVar Expr

subst :: Subst -> Expr -> Expr
subst phi e@(Meta i) = 
  case Map.lookup i phi of
    Just e' -> e'
    _ -> e
subst phi (Succ e) = Succ (subst phi e)
subst phi (App e es) = App (subst phi e) (map (subst phi) es)
subst phi (LLet x a b c) = LLet x (subst phi a) (subst phi b) (subst phi c)
subst phi (Pi dec x a b) = Pi dec x (subst phi a) (subst phi b)
subst phi (Lam x e)  = Lam x (subst phi e)
subst phi e = e

-- declarations --------------------------------------------------

data Declaration 
  = DataDecl Name Sized Co [Pos] Telescope Type [Constructor]  -- data/codata
  | FunDecl Co [(TypeSig,[Clause])]  -- mutual fun block / mutual cofun block
  | LetDecl Bool TypeSig Expr -- bool for eval ??
    deriving (Eq,Show)

data TypeSig = TypeSig { namePart :: Name, typePart :: Type }
             deriving (Eq,Show)

type Type = Expr

type Constructor = TypeSig

type TBind = (Dec,Name,Type)  -- same as TypeSig ??

type Telescope = [TBind]

data Clause = Clause [Pattern] Expr  
            deriving (Eq,Show)

data Pattern = VarP Name
             | ConP Co Name [Pattern]
             | SuccP Pattern
             | UnusableP Pattern  
{- a pattern which results from matching a coinductive type and
the corresponding size index is not in the coinductive result type of
the function.  Such a pattern is not usable for termination
checking. -}
             | DotP Expr
               deriving (Eq,Show)

type Case = (Pattern,Expr)

patternToExpr :: Pattern -> Expr
patternToExpr (VarP n)       = Var n
patternToExpr (ConP co n ps) = Con co n `App` (map patternToExpr ps)
patternToExpr (SuccP p)      = Succ (patternToExpr p)
patternToExpr (UnusableP p)  = patternToExpr p
patternToExpr (DotP e)       = e

completeP :: Pattern -> Bool
completeP (DotP _) = True
completeP (VarP _) = True
completeP (UnusableP p) = completeP p
completeP _ = False

isSuccessorPattern :: Pattern -> Bool
isSuccessorPattern (SuccP _) = True
isSuccessorPattern (DotP (Succ _)) = True
isSuccessorPattern _ = False

isVarP :: Pattern -> Bool
isVarP (VarP _) = True
isVarP (DotP (Var _)) = True
isVarP _ = False

shallowSuccP :: Pattern -> Bool
shallowSuccP (SuccP p) = isVarP p
shallowSuccP (DotP (Succ (Var _))) = True
shallowSuccP _ = False

rhsDefs :: [Clause] -> [Name]
rhsDefs cls = foldl (\ ns (Clause ps e) -> usedDefs e ++ ns) [] cls

teleToType' :: (Dec -> Dec) -> Telescope -> Type -> Type
teleToType' mod [] t = t
teleToType' mod ((dec,n,t):tel) t2 = Pi (mod dec) n t (teleToType' mod tel t2)

teleToType :: Telescope -> Type -> Type
teleToType = teleToType' id

teleToTypeErase :: Telescope -> Type -> Type
teleToTypeErase = teleToType' (\ dec -> dec { erased = True })

splitTeleType :: Int -> (Telescope,Type) -> (Telescope,Type)
splitTeleType 0 (tel,t) = (tel,t)
splitTeleType k (tel,(Pi dec n t t2)) = splitTeleType (k - 1) (tel ++ [(dec,n,t)],t2) 

typeToTele :: Type -> (Telescope, Type)
typeToTele t = ttt t []
    where 
      ttt :: Type -> Telescope -> (Telescope,Type)
      ttt (Pi dec n t' t2) tel = ttt t2 (tel ++ [(dec,n,t')])
      ttt x tel = (tel,x)                          
